## Licence: LGPL-v3
## Authors: Alexey Lindo (lindo@chalmers.se)
## TODO(alexey): Purpose of this program, input, output

IsOptimMesop <- function(measure, grad, tolerance.measure, tolerance.gradient){
    ## Checks if gradient non-negative and equals to zero within error tol on the support of measure
    ##
    ## Args:
    ##   measure: 
    ##   grad: 
    ##   tolerance.measure:
    ## tolerance.gradient:
    ##
    ## Returns:
    ##
    if (length(measure) != length(grad))
        stop("The lists should be of equal length")
    return(opt = all(grad >= -tolerance.gradient)
           && all(grad[measure >= tolerance.measure] <= tolerance.gradient))
}

PutMass <- function(measure, grad, mass, x, uni=FALSE){
    ## Puts mass to measure in the point x where gradient is minimal
    ## If uni=TRUE then we spread the mass uniformly among the points where the gradient is minimal
    ##
    ## Args:
    ##   measure:
    ##   mass:
    ##   x:
    ##   uni:
    ##
    ## Returns:
    ##
    if (length(measure) != length(grad))
        stop("The lists should be of equal length")
    if (uni){
        num <- length(grad[grad == grad[x]])  # the number of min elements in g
        measure[grad == grad[x]] <- measure[grad == grad[x]] + mass / num
    }
    else
        measure[x] <- measure[x] + mass
    return(measure)
}

TakeMass <- function(measure, mass, x){
    ## Takes mass from measure in the point x
    ##
    ## Args:
    ##   measure:
    ##   mass:
    ##   x:
    ##
    ## Returns:
    ##
    measure[x] <- measure[x] - mass  # taking mass from the first point where gradient is maximal
    return(measure)
}

MakeStep <-  function(measure, grad, step.size, uni, minim){
    ##
    ## Args:
    ##   measure:
    ##   grad:
    ##   step.size:
    ##   uni:
    ##   minim:
    ##
    ## Returns:
    ##
    if (!minim)
        grad <- -grad
    step.left <- step.size
    min.i <- which.min(grad)
    r <- which(grad >= abs(grad[min.i]))
    r <- r[(measure[r] > 0)]
    r.i <- order(grad[r], decreasing="TRUE")
    r <- r[r.i]
    if (grad[min.i] < 0){
        r <- append(r, which(grad == grad[min.i]))
    }
    while((step.left > 0) && (length(r) > 0)){
        if (grad[r[1]] < 0){
            measure <- PutMass(measure, grad, step.left, r[1], uni)
            step.left <- 0
        }
        else{
            measure <- TakeMass(measure, step.taken <- min(step.left, measure[r[1]]), r[1])
            step.left <- step.left - step.taken
            r <- r[-1]
        }
    }
    return(measure)
}

CalculateStep <- function(smallest.step.size,
                          space,
                          measure,
                          object.function,
                          gradient,
                          uni,
                          minim,
                          tol.measure,
                          tol.gradient,
                          dots){
    ##    
    ## Args:
    ##   smallest.step.size
    ##   space:
    ##   measure:
    ##   object.function:
    ##   gradient:
    ##   uni:
    ##   minim:
    ##   dots:
    ##
    ## Returns:
    ##
    step.size <- optimize(lower=0, upper=sum(measure), f=function(step.size){
                              return(do.call(object.function, c(list(MakeStep(measure,
                                                                              gradient,
                                                                              step.size,
                                                                              uni,
                                                                              minim),
                                                                     space),
                                                                dots)))}, tol=min(tol.measure, tol.gradient))$minim
    return(list(size = step.size,
                too.small = (step.size < smallest.step.size)))
}

CalculateStepArmijo <- function(smallest.step.size,
                                space,
                                measure,
                                object.function,
                                gradient,
                                uni,
                                minim,
                                alpha,
                                base,
                                dots){
    ##
    ## Args:
    ##   smallest.step.size
    ##   space:
    ##   measure:
    ##   object.function:
    ##   gradient:
    ##   uni:
    ##   minim:
    ##   alpha:
    ##   base:
    ##   sample:
    ##
    ## Returns:
    ##
    fval <-  do.call(object.function, c(list(measure, space), dots))
    rhs <- sum(gradient * measure)
    anm <- alpha * rhs
    armijo.condition <- FALSE
    too.small <- FALSE
    base.new <- base
    while(!armijo.condition && !too.small){
        if (base.new < smallest.step.size){
            too.small <- TRUE
        }
        measure.new <- MakeStep(measure,
                                gradient,
                                base.new,
                                uni,
                                minim)
        fval.new <- do.call(object.function, c(list(measure.new, space), dots))
        rhs <- sum(gradient * measure.new)
        armijo.condition <- ((fval.new - fval) <=  (alpha * rhs - anm))
        if (!armijo.condition && !too.small){
            base.new <- base.new * base
        }
    }
    return(list(size = base.new,
                too.small = too.small))

}

setMethod("initialize",
          "MesOp",
          function(.Object,
                   number.of.steps = 1,
                   ## key paramenters
                   space = NULL,
                   measure = NULL,
                   object.function = NULL,
                   gradient.function = NULL,
                   ## parameters of the numerical method
                   file.name = NA_character_,
                   minim = TRUE,
                   smallest.step.size = as.double(10^(-6)),
                   step.algo = as.integer(0),
                   step.size = 0.1 * sum(measure) + 10 * smallest.step.size, # TODO(alexey): Fix this
                   tolerance.measure = as.double(10^{-6}),
                   tolerance.gradient = as.double(10^{-6}),
                   uni = FALSE,
                   ## visual parameters
                   details = FALSE,
                   visualize = FALSE,
                   ## internals
                   alpha = 0.7,
                   base = 0.5,
                   true.measure = NA,
                   ## parameters of the Object and Gradient funcitons.
                   ...){
              ## key parameters
              .Object@space = space
              .Object@measure = measure
              .Object@object.function = object.function
              .Object@gradient.function = gradient.function
              ## parameters of the Object and Gradient functions.
              .Object@dots = list(...)
              ## parameters of the numerical method
              .Object@file.name = file.name
              .Object@minim = minim
              .Object@smallest.step.size = smallest.step.size
              .Object@step.algo = step.algo
              .Object@step.size = step.size
              .Object@tolerance.measure = tolerance.measure
              .Object@tolerance.gradient = tolerance.gradient
              .Object@uni = uni
              ## visual parameters
              .Object@details = details
              .Object@visualize = visualize
              ## internals
              .Object@alpha = alpha
              .Object@base = base
              return(GoSteep(.Object, number.of.steps))
          }
          )

setGeneric(name = "GoSteep", function(object, number.of.steps, ...){standardGeneric("GoSteep")})

setMethod(f="GoSteep", signature="MesOp", definition=function(object, number.of.steps = 1, ...){
              ##  
              ##
              ## Args:
              ##   number.of.steps:
              ##   object:
              ##  
              ## Returns:
              ##
              ## Now we are trying the get the value of the goal function
              fval <- do.call(object@object.function, c(list(object@measure, object@space), object@dots))
              ## Now we are trying to get the value of the gradient on the space
              object@gradient <- do.call(object@gradient.function, c(list(object@measure, object@space), object@dots))
              object@optim <- IsOptimMesop(object@measure,
                                           object@gradient,
                                           object@tolerance.measure,
                                           object@tolerance.gradient)

              if (!object@optim){
                  cat("Optimizing... \n")
                  i <- 1
                  while ((i <= number.of.steps) & !object@optim & !object@too.small.step.size){
                      if (object@step.algo == 1){
                          x <- CalculateStep(object@smallest.step.size,
                                             object@space,
                                             object@measure,
                                             object@object.function,
                                             object@gradient,
                                             object@uni,
                                             object@minim,
                                             object@tolerance.measure,
                                             object@tolerance.gradient,
                                             object@dots)
                      }
                      else{
                          x <- CalculateStepArmijo(object@smallest.step.size,
                                                   object@space,
                                                   object@measure,
                                                   object@object.function,
                                                   object@gradient,
                                                   object@uni,
                                                   object@minim,
                                                   object@alpha,
                                                   object@base,
                                                   object@dots)
                      }
                      step.size.old <- object@step.size
                      object@step.size <- x$size
                      object@too.small.step.size <- x$too.small
                      object@measure <- MakeStep(object@measure,
                                                 object@gradient,
                                                 object@step.size,
                                                 object@uni,
                                                 object@minim)
                      fval.old <- fval
                      fval <-  do.call(object@object.function, c(list(object@measure, object@space), object@dots))
                      object@gradient <- do.call(object@gradient.function, c(list(object@measure, object@space), object@dots))

                      ## Printing step information
                      PrintStepInformation(i,
                                           fval.old,
                                           fval,
                                           step.size.old,
                                           object@step.size,
                                           object@details)
                      if (object@visualize)
                          Draw(object@space,
                               object@measure,
                               object@gradient,
                               object@minim)
                      i <-  i + 1
                      object@steps.done <- object@steps.done + as.integer(1)
                      object@optim <- IsOptimMesop(object@measure,
                                                   object@gradient,
                                                   object@tolerance.measure,
                                                   object@tolerance.gradient)
                  }
              }

              ## Printing summary information
              PrintSummary(object@space,
                           object@measure,
                           object@optim,
                           object@tolerance.measure,
                           object@step.size,
                           object@too.small.step.size,
                           object@details)
              
              return(object)
          })


PrintStepInformation <- function(step.number,
                                 fval.old,
                                 fval,
                                 step.size.old,
                                 step.size,
                                 details){
                                        #
                                        #
    ## Args:
    ## step.number:
    ##   fval.old:
    ##   fval:
    ##   step.size.old:
    ##   step.size:
    ##   details:
    ##
    ## Returns:
    ##
    cat("\nStep number:", step.number, "\n")
    if (!details){
        if (fval.old < fval)
            cat("(f+,")
        else{
            if (fval.old > fval)
                cat("(f-,")
            else
                cat("(f.,")
        }  
        if (step.size > step.size.old)
            cat("st+)\n")
        else{
            if (step.size < step.size.old)
                cat("st-)\n")
            else
                cat("st.)\n")
        }
    }
    else{
        cat("The value of Object function was:", fval.old, "the new values is:", fval, "\n")
        cat("The step size was:", step.size.old, "the current step size is:", step.size, "\n")
    }
}

LocateAtoms <- function(space, measure, tolerance.measure){
    ##
    ##
    ## Args:
    ##   space:
    ##   measure:
    ##   tolerance.measure
    ##
    ## Returns:
    ##

    ## If the measure of the point is less then tol.supp then we decide that measure of this point equals to zero
    tol.supp <- (tolerance.measure / length(measure)) * sum(measure)

    ## Now generate string to print measure in its support points only
    supp <- space[measure > tol.supp]
    atoms <- measure[measure > tol.supp]
    number.of.atoms <- length(atoms)

    return(list(supp = supp,
                atoms = atoms,
                number.of.atoms = number.of.atoms))
}

PrintMeasureOnSupport <- function(space, measure, tolerance.measure){
    ##
    ##
    ## Args:
    ##   space:
    ##   measure:
    ##   tolerance.measure:
    ##
    ## Returns:
    ##
    x <- LocateAtoms(space, measure, tolerance.measure)
    if (x$number.of.atoms == 0)
        cat("The obtained measure contains no atoms\n\n")
    else{
        if (x$number.of.atoms == 1)
            cat("The obtained measure contains one atom:\n\n")
        else{
            cat("The obtained measure contains ", x$number.of.atoms, "atoms \n(you might consider collapsing close ones)\n\n")
        }
    }
    if (length(x$supp) > 0){
        if (length(x$supp) < 31){
            data <- matrix(c(x$supp, x$atoms), ncol = 2)
            colnames(data) <- c("Support", "Atoms")
            rownames(data) <- 1:length(x$supp)
            print(data)
        }
        else{
            output.limits <- c(1:15)
            data <- matrix(c(x$supp[output.limits], x$atoms[output.limits]), ncol = 2)
            colnames(data) <- c("Support", "Atoms")
            rownames(data) <- output.limits
            print(data)
            cat("...\n")
            output.limits <- c((length(x$supp) - 15):length(x$supp))
            data <- matrix(c(x$supp[output.limits], x$atoms[output.limits]), ncol = 2)
            colnames(data) <- c("Support", "Atoms")
            rownames(data) <- output.limits
            print(data)
        }
    }
}

PrintSummary <- function(space,
                         measure,
                         optim,
                         tolerance.measure,
                         step.size,
                         too.small.step.size,
                         details){
    ##
    ##
    ## Args:
    ##   space:
    ##   measure:
    ##   optim:
    ##   tolerance.measure:
    ##   step.size:
    ##   too.small.step.size:
    ##
    ## Returns:
    ##

    if (optim)
        cat("\n Optimum is attained\n\n")
    else{
        if (too.small.step.size)
            cat("\n Computation stopped, too small step size:", step.size, "\n\n")
        else{
            cat("\n Optimum is not attained\n")
            cat("\n Consider rerunning GoSteep() and possibly change the parameters\n\n")
        }
    }
    if (details)
        PrintMeasureOnSupport(space, measure, tolerance.measure)
}

setMethod("print",
          signature(x = "MesOp"),
          definition = function(x){
              ##
              ##
              ## Args:
              ##   x:
              ##
              ## Returns:
              ##
              if (x@minim)
                  cat("\nMinimization of measure dependent functional\n\n")
              else
                  cat("\nMaximization of measure dependent functional\n\n")
              cat("Value of the object function on the measure below:", x@last.fval, "\n")
                                        # length(x@space) == length(x@measure)
                                        # length(x@gradient) == length(x@space)
                                        # the last element of x@gradient is the value of the derivative with respect to real parameter
              if (length(x@space) < 31){
                  data <- matrix(c(x@space, x@measure, x@gradient[-length(gradient)]), ncol = 3)
                  colnames(data) <- c("Space", "Measure", "Gradient")
                  rownames(data) <- 1:length(space)
                  print(data)
              }
              else{
                  output.limits <- c(1:15)
                  data <- matrix(c(x@space[output.limits], x@measure[output.limits], x@gradient[output.limits]), ncol = 3)
                  colnames(data) <- c("Space", "Measure", "Gradient")
                  rownames(data) <- output.limits
                  print(data)
                  cat("...\n")
                  output.limits <- c((length(x@space) - 15):length(x@space))
                  data <- matrix(c(x@space[output.limits], x@measure[output.limits], x@gradient[output.limits]), ncol = 3)
                  colnames(data) <- c("Space", "Measure", "Gradient")
                  rownames(data) <- output.limits
                  print(data)
              }
          })

setMethod("summary",
          signature(object = "MesOp"),
          definition = function(object){
              ##
              ##
              ## Args:
              ##   object:
              ##
              ## Returns:
              ##
              if (object@minim)
                  ## TODO(alexey): Uncomment when object.function.name will be added
                  ## cat("\nMinimization of function", object@object.function.name, "\n")
                  cat("\nMinimization of measure dependent functional\n")
              else
                  ## TODO(alexey): Uncomment when object.function.name will be added
                  ## cat("\nMaximization of function", deparse(substitute(object@object.function)), ".\n")
                  cat("\nMaximization of measure dependent functional\n")
              ## TODO(alexey): Uncomment when gradient.function.name will be added
              ## cat("with gradient function", deparse(substitute(object@gradient.function)), ".\n")
              ## TODO(alexey): Add algorithms names for object@step.algo output
              ## cat("\nAlgorithm", object@step.algo, "was called with tolerance", object@tolerance.measure, "for measure, and", object@tolerance.gradient,  "for gradient \n")
              if (object@optim){
                  cat("\n", "Optimum is attained.", "\n")
                  PrintMeasureOnSupport(object@space, object@measure, object@tolerance.measure)
              }
              else{
                  cat("\n", "Optimum is not attained.", "\n")
                  PrintMeasureOnSupport(object@space, object@measure, object@tolerance.measure)
              }

              ## Now print some statistics
              ## TODO(alexey): Uncomment when object.function.name will be added
              ## cat("\n\nValue of the objective function: ", object@object.function.name, ":", object@last.fval, "\n")
              ## cat("\nMininam value of gradient function: ", mesop$var.grad)
              cat("\nTotal number of steps made: ", object@steps.done, "\n")
              invisible(object)
          })

Draw <- function(space,
                 measure,
                 grad,
                 tolerance.measure,
                 minim=TRUE){
    ##
    ## Args:
    ##   space:
    ##   measure:
    ##   grad:
    ##   minim:
    ##
    ## Returns:
    ##
    t0 <- (tolerance.measure * sum(measure) / length(measure))
    t.a <- max(grad) - min(grad)
    t.o <- max(grad[measure >= t0])
    if (t.a == 0){
        t.div <- 0
    }
    else {
        t.div <- t.o / t.a
    }
    par(mfrow=c(2,1))
    plot(space,
         grad,
         type = "l",
         main = "Gradient function",
         lwd = 2,
         axes = TRUE,
         col = "darkred",
         xlab = "",
         ylab = "")
    abline(h=0, col = "green")
    if (minim)
        abline(h = min(grad), lwd = 2, col = "green")
    else
        abline(h = max(grad), lwd = 2, col = "green")
    abline(h = -t.div, lwd = 1, col = "blue")
    plot(space,
         measure,
         type = "h",
         lwd = 2,
         main = "Measure",
         axes = TRUE,
         col = "red",
         xlab = "",
         ylab = "")
    abline(h=0)
}

setMethod("plot",
          signature(x = "MesOp"),
          definition = function(x){
              ##
              ##
              ## Args:
              ##   x:
              ##
              ## Returns:
              ##
              ## TODO(alexey): Fix the case when x@grad is NULL
              Draw(x@space,
                   x@measure,
                   x@gradient,
                   x@tolerance.measure,
                   x@minim)
          })


